# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['reportseff']

package_data = \
{'': ['*']}

install_requires = \
['click>=6.7']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.8.2,<5.0.0']}

entry_points = \
{'console_scripts': ['reportseff = reportseff.console:main']}

setup_kwargs = {
    'name': 'reportseff',
    'version': '2.4.2',
    'description': 'Tablular seff output',
    'long_description': '[![Tests](https://github.com/troycomi/reportseff/workflows/Tests/badge.svg)](https://github.com/troycomi/reportseff/actions?workflow=Tests)\n[![codecov](https://codecov.io/gh/troycomi/reportseff/branch/main/graph/badge.svg)](https://codecov.io/gh/troycomi/reportseff)\n[![PyPI](https://img.shields.io/pypi/v/reportseff.svg)](https://pypi.org/project/reportseff/)\n\n# `reportseff`\n\n> A python script for tabular display of slurm efficiency information\n\n![Example](https://github.com/troycomi/reportseff/raw/main/imgs/example.png)\n\n## About\n\n### Motivation\n\nWhether a sys admin or cluster user, knowing how well you are estimating job\nresources can help streamline job scheduling and maximize your priority. If you\nhave ever tried to use `sacct` you probably had some trouble interpreting the\noutput.  While `seff` or `jobstats` can provide detailed summaries, they don\'t\nscale easily to array jobs or offer a way to see all the jobs from a single\nuser.  `reportseff` aims to fill this role.  Read more about the [motivation\nfor reportseff](https://github.com/troycomi/reportseff/blob/main/ABOUT.md).\n\n### Audience\n\nIf you are running more than one slurm job at a time, you should try\n`reportseff`.  Users of HPC systems can get an idea how well they estimate\nresource usage.  By tuning these values, you can get scheduled earlier and not\nbe penalized for unused allocations.  Since `reportseff` can parse job ids from\nslurm output files, it simplifies the task of identifying which jobs have\nfailed and why. Sys admins can pipe `reportseff` output to identify users with\npoor utilization or produce summaries at the end of a billing cycle.\n\n### Implementation\n\n`reportseff` is a wrapper around `sacct` that provides more complex option\nparsing, simpler options, and cleaner, colored outputs.  All querying is\nperformed in a single call to `sacct` and should have similar performance.\nMulti-node and GPU utilization is acquired from information contained in the\n`AdminComment` field, as generated by `jobstats`.\n\n## Usage\n\n### Installation\n\n`reportseff` runs on python >= 3.6.\nThe only external dependency is click (>= 6.7).\nCalling\n\n```sh\npip install --user reportseff\n# OR\npipx install reportseff\n```\n\nwill create command line bindings and install click.\n\n### Sample Usage\n\nTry `reportseff -u $USER` or just `reportseff` in a directory with some slurm\noutputs.  You may be surprised by your results!\n\n#### Single job\n\nCalling `reportseff` with a single jobid will provide equivalent information to\nseff for that job. `reportseff 24371789` and `reportseff map_fastq_24371789`\nproduce the following output:\n\n```txt\n   JobID      State          Elapsed   CPUEff   MemEff\n24371789    COMPLETED       03:08:03   71.2%    45.7%\n```\n\n#### Single array job\n\nProviding either the raw job id or the array job id will get efficiency\ninformation for a single element of the array job. `reportseff 24220929_421`\nand `reportseff 24221219` generate:\n\n```txt\n       JobID      State          Elapsed    CPUEff   MemEff\n24220929_421    COMPLETED       00:09:34    99.0%    34.6%\n```\n\n#### Array job group\n\nIf the base job id of an array is provided, all elements of the array will\nbe added to the output. `reportseff 24220929`\n\n```txt\n          JobID      State          Elapsed    CPUEff   MemEff\n     24220929_1    COMPLETED       00:10:43    99.2%    33.4%\n    24220929_11    COMPLETED       00:10:10    99.2%    37.5%\n    24220929_21    COMPLETED       00:09:25    98.8%    36.1%\n    24220929_31    COMPLETED       00:09:19    98.9%    33.3%\n    24220929_41    COMPLETED       00:09:23    98.9%    33.3%\n    24220929_51    COMPLETED       00:08:02    98.5%    36.3%\n    ...\n   24220929_951    COMPLETED       00:25:12    99.5%    33.5%\n   24220929_961    COMPLETED       00:39:26    99.7%    34.1%\n   24220929_971    COMPLETED       00:24:11    99.5%    34.2%\n   24220929_981    COMPLETED       00:24:50    99.5%    44.3%\n   24220929_991    COMPLETED       00:13:05    98.7%    33.7%\n```\n\n#### Glob expansion of slurm outputs\n\nBecause slurm output files can act as job id inputs, the following can\nget all seff information for a given job name:\n\n```txt\nslurm_out  ❯❯❯ reportseff split_ubam_24*\n              JobID      State          Elapsed   CPUEff   MemEff\nsplit_ubam_24342816    COMPLETED       23:30:32   99.9%    4.5%\nsplit_ubam_24342914    COMPLETED       22:40:51   99.9%    4.6%\nsplit_ubam_24393599    COMPLETED       23:43:36   99.4%    4.4%\nsplit_ubam_24393655    COMPLETED       21:36:58   99.3%    4.5%\nsplit_ubam_24418960     RUNNING        02:53:11    ---      ---\nsplit_ubam_24419972     RUNNING        01:26:26    ---      ---\n```\n\n#### No arguments\n\nWithout arguments, reportseff will try to find slurm output files in the\ncurrent directory. Combine with `watch` to monitor job progress:\n`watch -cn 300 reportseff --modified-sort`\n\n```txt\n                JobID           State          Elapsed   CPUEff   MemEff\n   split_ubam_24418960          RUNNING        02:56:14    ---      ---\nfastq_to_ubam_24419971          RUNNING        01:29:29    ---      ---\n   split_ubam_24419972          RUNNING        01:29:29    ---      ---\nfastq_to_ubam_24393600         COMPLETED     1-02:00:47   58.3%    41.1%\n    map_fastq_24419330          RUNNING        02:14:53    ---      ---\n    map_fastq_24419323          RUNNING        02:15:24    ---      ---\n    map_fastq_24419324          RUNNING        02:15:24    ---      ---\n    map_fastq_24419322          RUNNING        02:15:24    ---      ---\nmark_adapters_24418437         COMPLETED       01:29:23   99.8%    48.2%\nmark_adapters_24418436         COMPLETED       01:29:03   99.9%    47.4%\n```\n\n#### Filtering slurm output files\n\nOne useful application of `reportseff` is filtering a directory of slurm output\nfiles based on the state or time since running. Additionally, if only the\n`jobid` is specified as a format output, the filenames will be returned in a\npipe-friendly manner:\n\n```txt\nold_runs   ❯❯❯ reportseff --since d=4 --state Timeout\n\n                 JobID   State      Elapsed  CPUEff   MemEff\ncall_variants_31550458  TIMEOUT    20:05:17  99.5%     0.0%\ncall_variants_31550474  TIMEOUT    20:05:17  99.6%     0.0%\ncall_variants_31550500  TIMEOUT    20:05:08  99.7%     0.0%\nold_runs   ❯❯❯ reportseff --since d=4 --state Timeout --format jobid\ncall_variants_31550458\ncall_variants_31550474\ncall_variants_31550500\n```\n\nTo find all lines with `output:` in jobs which have timed out or failed\nin the last 4 days:\n\n```sh\nreportseff --since \'d=4\' --state TO,F --format jobid | xargs grep output:\n```\n\n### Arguments\n\nJobs can be passed as arguments in the following ways:\n\n- Job ID such as 1234567.  If the id is part of an array job, only the element\nfor that ID will be displayed.  If the id is the base part of an array job,\nall elements in the array will be displayed.\n- Array Job ID such as 1234567\\_89.  Will display only the element specified.\n- Slurm output file.  Format must be BASE\\_%A\\_%a.  BASE is optional as is a\n\'.out\' suffix.  Unix glob expansions can also be used to filter which jobs\nare displayed.\n- From current directory.  If no argument is supplied, `reportseff` will attempt\nto find slurm output files in the current directory as described above.\nIf a user is provided, instead `reportseff` will show recent jobs for that user.\nIf only `since` is set, all recent jobs for all users will be shown (if allowed).\n- Supplying a directory as a single argument will override the current\ndirectory to check for slurm outputs.\n\n### Options\n\n- `--color/--no-color`: Force color output or not.  By default, will force color\n  output.  With the no-color flag, click will strip color codes for everything\n  besides stdout.\n- `--modified-sort`: Instead of sorting by filename/jobid, sort by last\n  modification time of the slurm output file.\n- `--debug`: Write sacct result to stderr.\n- `--user/-u`: Ignore job arguments and instead query sacct with provided user.\n  Returns all jobs from the last week.\n- `--state/-s`: Output only jobs with states matching one of the provided options.\n  Accepts comma separated values of job codes (e.g. \'R\') or full names\n  (e.g. RUNNING).  Case insensitive.\n- `--not-state/-S`: Output only jobs with states not matching any of the provided options.\n  Accepts comma separated values of job codes (e.g. \'R\') or full names\n  (e.g. RUNNING).  Case insensitive.\n- `--format`: Provide a comma separated list of columns to produce. Prefixing the\n  argument with `+` adds the specified values to the defaults.  Values can\n  be any valid column name to sacct and the custom efficiency values: TimeEff,\n  cpuEff, MemEff.  Can also optionally set alignment (<, ^, >) and maximum width.\n  Default is center-aligned with a width of the maximum column entry.  For\n  example, `--format \'jobid%>,state%10,memeff%<5\'` produces 3 columns with:\n  - JobId aligned right, width set automatically\n  - State with width 10 (center aligned by default)\n  - MemEff aligned left, width 5\n- `--since`: Limit results to those occurring after the specified time.  Accepts\n  sacct formats and a comma separated list of key/value pairs.  To get jobs in\n  the last hour and a half, can pass `h=1,m=30`.\n- `--node/-n`: Display information for multi-node jobs; requires additional\n  sacct fields from jobstats.\n- `--node-and-gpu/-g`: Display information for multi-node jobs and GPU information;\n  requires additional sacct fields from jobstats.\n- `--parsable/-p`: Ignore formatting and output as a `|` delimited table.  Useful\n  for piping into more complex analyses.\n\n## Status, Contributions, and Support\n\n`reportseff` is actively maintained but currently feature complete.  If there\nis a function missing, please open an issue to discuss its merit!\n\nBug reports, pull requests, and any feedback are welcome! Prior to submitting\na pull request, be sure any new features have been tested and all unit tests\nare passing. In the cloned repo with\n[poetry](https://github.com/python-poetry/poetry#installation) installed:\n\n```sh\npoetry install\npoetry run pytest\npoetry run pre-commit install\nnox\n```\n\n## Troubleshooting\n\n### I can\'t install, what is pip?\n\n[pip](https://pip.pypa.io/en/stable/) is the package installer for python.  If\nyou get an error that pip isn\'t found, look for a python/anaconda/conda module.\n[pipx](https://pypa.github.io/pipx/) ensures that each application is installed\nin an isolated environment.  This resolves issues of dependency versions and\nallows applications to be run from any environment.\n\n### I get an error about broken pipes when chaining to other commands\n\nPython will report that the consumer of process output has closed the stream\n(i.e. the pipe) while still attempting to write.  Newer versions of click\nshould suppress the warning output, but it seems to not always work.  Besides\nsome extra printing on stderr, the output is not affected.\n\n### My jobs don\'t have any information about multiple nodes or GPU efficiency\n\nBecause `sacct` doesn\'t currently record this information, `reportseff`\nretrieves it from a custom field from `jobstats`, developed at Princeton\nUniversity.  If you are outside a Research Computing cluster, that information\nwill likely be absent.  Node-level reporting is only shown for jobs which use\nmultiple nodes or GPUs.  If you need a list of where jobs were run, you can add\n`--format +NodeList`.\n\n### My output is garbled with ESC[0m all over, where\'s the color?\n\nThose are ANSI color codes.  Click will usually strip these if it detects\nthe consuming process can\'t display color codes, but `reportseff` defaults\nto always display them.  If you don\'t care for color, use the `--no-color`\noption.  For less, you can set\n```\nexport LESS="-R"\n```\nin your `.bashrc`, or just type `-R` in an active less process.  Some versions\nof `watch` require the `-c` option to display color, others can\'t display\ncolors properly.  If you search for `ansi color <tool>` you should get some\nsolutions.\n\n## Acknowledgments\n\nThe code for calling sacct and parsing the returning information was taken\nfrom [Slurmee](https://github.com/PrincetonUniversity/slurmee).\n\nStyle and tooling from [hypermodern python](https://cjolowicz.github.io/posts/hypermodern-python-01-setup/)\n\nCode review provided from a [repo-review](https://researchcomputing.princeton.edu/services/repo-review-consultations)\nwhich vastly improved this readme.\n',
    'author': 'Troy Comi',
    'author_email': 'troycomi@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/troycomi/reportseff',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0',
}


setup(**setup_kwargs)
